/*
 #
 #  File        : captcha.cpp
 #                ( C++ source file )
 #
 #  Description : Captcha images generator.
 #                This file is a part of the CImg Library project.
 #                ( http://cimg.sourceforge.net )
 #
 #  Copyright   : David Tschumperle
 #                ( http://www.greyc.ensicaen.fr/~dtschump/ )
 #
 #  License     : CeCILL v2.0
 #                ( http://www.cecill.info/licences/Licence_CeCILL_V2-en.html )
 #
 #  This software is governed by the CeCILL  license under French law and
 #  abiding by the rules of distribution of free software.  You can  use,
 #  modify and/ or redistribute the software under the terms of the CeCILL
 #  license as circulated by CEA, CNRS and INRIA at the following URL
 #  "http://www.cecill.info".
 #
 #  As a counterpart to the access to the source code and  rights to copy,
 #  modify and redistribute granted by the license, users are provided only
 #  with a limited warranty  and the software's author,  the holder of the
 #  economic rights,  and the successive licensors  have only  limited
 #  liability.
 #
 #  In this respect, the user's attention is drawn to the risks associated
 #  with loading,  using,  modifying and/or developing or reproducing the
 #  software by the user in light of its specific status of free software,
 #  that may mean  that it is complicated to manipulate,  and  that  also
 #  therefore means  that it is reserved for developers  and  experienced
 #  professionals having in-depth computer knowledge. Users are therefore
 #  encouraged to load and test the software's suitability as regards their
 #  requirements in conditions enabling the security of their systems and/or
 #  data to be ensured and,  more generally, to use and operate it in the
 #  same conditions as regards security.
 #
 #  The fact that you are presently reading this means that you have had
 #  knowledge of the CeCILL license and that you accept its terms.
 #
*/

#ifndef cimg_debug
#define cimg_debug 1
#endif
#include "../CImg.h"
using namespace cimg_library;

// The lines below are necessary when using a non-standard compiler such as visualcpp6.
#ifdef cimg_use_visualcpp6
#define std
#endif
#ifdef min
#undef min
#undef max
#endif

// Start main procedure
//----------------------
int main(int argc,char **argv) {

  // Read command line parameters
  //------------------------------
  cimg_usage("Captcha generator");
  const char *input_text   = cimg_option("-i",(char*)0,"Input captcha text (default is random)");
  const bool force_numbers = cimg_option("-n",false,"Put only numbers in generated captcha");
  const char *file_o       = cimg_option("-o",(char*)0,"Output image file");
  const bool add_border    = cimg_option("-b",true,"Add border to captcha image");
  const bool visu          = cimg_option("-visu",true,"Enable visualization if no output file");

  // Generate random captcha text (6 char max).
  //------------------------------
  char captcha_text[7] = { 0 };
  if (input_text) std::strncpy(captcha_text,input_text,6);
  else {
    cimg::srand();
    for (unsigned int k=0; k<6; ++k) {
      const int
        v = (int)std::rand()%(force_numbers?1:2),
        val = (int)(v==0?('1'+(std::rand()%9)):('A'+(std::rand()%26)));
      if (val=='1' || val=='L' || val=='7' ||
          val=='0' || val=='O' || val=='Q' ||
          val=='8' || val=='B') --k;
      else captcha_text[k] = val;
    }
  }

  // Create captcha image
  //----------------------

  // Write colored and distorted text
  CImg<unsigned char> captcha(256,64,1,3,0), color(3);
  char letter[2] = { 0 };
  for (unsigned int k=0; k<6; ++k) {
    CImg<unsigned char> tmp;
    *letter = captcha_text[k];
    if (*letter) {
      { cimg_forX(color,i) color[i] = (unsigned char)(128+(std::rand()%127)); }
      tmp.draw_text(letter,(int)(2+8*cimg::rand()),(int)(12*cimg::rand()),color.ptr(),0,std::rand()%2?38:57,1);
      const unsigned int dir = std::rand()%4, wph = tmp.dimx()+tmp.dimy();
      { cimg_forXYV(tmp,x,y,v) {
        const int val = dir==0?x+y:(dir==1?x+tmp.dimy()-y:(dir==2?y+tmp.dimx()-x:tmp.dimx()-x+tmp.dimy()-y));
        tmp(x,y,v) = (unsigned char)cimg::max(0.0f,cimg::min(255.0f,1.5f*tmp(x,y,v)*val/wph));
      }}
      if (std::rand()%2) tmp = (tmp.get_dilate(3)-=tmp);
      tmp.blur((float)cimg::rand()*0.8f).normalize(0,255);
      const float sin_offset = (float)cimg::crand()*3, sin_freq = (float)cimg::crand()/7;
      { cimg_forYV(captcha,y,v) captcha.get_shared_line(y,0,v).translate((int)(4*std::cos(y*sin_freq+sin_offset))); }
      captcha.draw_image(tmp,6+40*k,0);
    }
  }

  // Add geometric and random noise
  CImg<unsigned char> copy = captcha.get_fill(0);
  for (unsigned int l=0; l<3; ++l) {
    if (l) copy.blur(0.5f).normalize(0,148);
    for (unsigned int k=0; k<10; ++k) {
      cimg_forX(color,i) color[i] = (unsigned char)(128+cimg::rand()*127);
      if (cimg::rand()<0.5f) copy.draw_circle((int)(cimg::rand()*captcha.dimx()),
                                              (int)(cimg::rand()*captcha.dimy()),
                                              (int)(cimg::rand()*30),
                                              color,0.6f,~0UL);
      else copy.draw_line((int)(cimg::rand()*captcha.dimx()),
                          (int)(cimg::rand()*captcha.dimy()),
                          (int)(cimg::rand()*captcha.dimx()),
                          (int)(cimg::rand()*captcha.dimy()),
                          color,0.6f);
    }
  }
  captcha|=copy;
  captcha.noise(10,2);

  if (add_border) captcha.draw_rectangle(0,0,captcha.dimx()-1,captcha.dimy()-1,CImg<unsigned char>::vector(255,255,255),1.0f,~0UL);
  captcha = captcha.get_fill(255) - captcha;

  // Write output image and captcha text
  //-------------------------------------
  std::printf("%s\n",captcha_text);
  if (file_o) captcha.save(file_o);
  else if (visu) {
    CImgDisplay disp(CImg<unsigned char>(512,128,1,3,180).draw_image(captcha,128,32),captcha_text,0);
    while (!disp.is_closed && !disp.key) { disp.wait(); if (disp.is_resized) disp.resize(disp); }
  }
  return 0;
}
