/*
 #
 #  File        : inrcast.cpp
 #                ( C++ source file )
 #
 #  Description : 1D/2D/3D Image converter and visualizer based on command line arguments.
 #                This file is a part of the CImg Library project.
 #                ( http://cimg.sourceforge.net )
 #
 #  Copyright   : David Tschumperle
 #                ( http://www.greyc.ensicaen.fr/~dtschump/ )
 #
 #  License     : CeCILL v2.0
 #                ( http://www.cecill.info/licences/Licence_CeCILL_V2-en.html )
 #
 #  This software is governed by the CeCILL  license under French law and
 #  abiding by the rules of distribution of free software.  You can  use,
 #  modify and/ or redistribute the software under the terms of the CeCILL
 #  license as circulated by CEA, CNRS and INRIA at the following URL
 #  "http://www.cecill.info".
 #
 #  As a counterpart to the access to the source code and  rights to copy,
 #  modify and redistribute granted by the license, users are provided only
 #  with a limited warranty  and the software's author,  the holder of the
 #  economic rights,  and the successive licensors  have only  limited
 #  liability.
 #
 #  In this respect, the user's attention is drawn to the risks associated
 #  with loading,  using,  modifying and/or developing or reproducing the
 #  software by the user in light of its specific status of free software,
 #  that may mean  that it is complicated to manipulate,  and  that  also
 #  therefore means  that it is reserved for developers  and  experienced
 #  professionals having in-depth computer knowledge. Users are therefore
 #  encouraged to load and test the software's suitability as regards their
 #  requirements in conditions enabling the security of their systems and/or
 #  data to be ensured and,  more generally, to use and operate it in the
 #  same conditions as regards security.
 #
 #  The fact that you are presently reading this means that you have had
 #  knowledge of the CeCILL license and that you accept its terms.
 #
*/

#define cimg_debug 1
#include "../CImg.h"
using namespace cimg_library;
// The undef below is necessary when using a non-standard compiler.
#ifdef cimg_use_visualcpp6
#define std
#endif
#ifdef min
#undef min
#undef max
#endif

//---------------------------------------------
// Convert to output image, save and visualize
//---------------------------------------------
template<typename T,typename t> void output_file(int argc, char **argv, CImgList<T>& src,
                                                 const CImgList<char>& filename, const bool display_help, const t&) {

  // Read command line arguments
  if (display_help) cimg_help("\n Parameters : Input data"
                              "\n -----------------------");
  static const bool a_dimensions = cimg_option("-dim",false,"Print image dimensions and exit");
  static const char* typei       = cimg_option("-i","float","Set input type of image pixels "
                                               "(char,uchar,ushort,short,uint,int,ulong,long,float,double)");
  static const bool a_endian     = cimg_option("-endian",false,"Invert endianness of pixel buffer");
  if (display_help) {
    cimg_option("-g0","352x288","Dimensions of input data (for YUV,RAW,RGB,RGBA,CIMG files)");
    cimg_option("-frames",(char*)0,"Video files : Range 'first,last,step' of video frame(s) to read");
    cimg_option("-multiplexed",false,"RAW files : Use multiplexed mode");
    cimg_option("-endian",false,"RAW files : Set big/little endian mode");
    cimg_option("-bgcolor","32,64,96","OFF files : Default background color");
    cimg_option("-render",4,"OFF files : Render type for static objects");
    cimg_option("-renderm",-1,"OFF files : Render type for moving objects");
    cimg_option("-of",false,"OFF files : Consider oriented faces");
    cimg_option("-if",false,"OFF files : Invert face orientations");
  }

  if (display_help) cimg_help("\n Parameters : Color space conversion"
                              "\n -----------------------------------");
  static const bool a_rgb2hsv     = cimg_option("-rgb2hsv",false,"from RGB to HSV");
  static const bool a_rgb2hsl     = cimg_option("-rgb2hsl",false,"from RGB to HSL");
  static const bool a_rgb2hsi     = cimg_option("-rgb2hsi",false,"from RGB to HSI");
  static const bool a_rgb2yuv     = cimg_option("-rgb2yuv",false,"from RGB to YUV");
  static const bool a_rgb2ycbcr   = cimg_option("-rgb2ycbcr",false,"from RGB to YCbCr");
  static const bool a_rgb2xyz     = cimg_option("-rgb2xyz",false,"from RGB to XYZ");
  static const bool a_rgb2lab     = cimg_option("-rgb2lab",false,"from RGB to Lab");
  static const bool a_rgb2cmy     = cimg_option("-rgb2cmy",false,"from RGB to CMY");
  static const bool a_rgb2cmyk    = cimg_option("-rgb2cmyk",false,"frome RGB to CMYK");
  static const bool a_cmyk2rgb    = cimg_option("-cmyk2rgb",false,"frome CMYK to RGB");
  static const bool a_cmy2rgb     = cimg_option("-cmy2rgb",false,"frome CMY to RGB");
  static const bool a_lab2rgb     = cimg_option("-lab2rgb",false,"from Lab to RGB");
  static const bool a_xyz2rgb     = cimg_option("-xyz2rgb",false,"from XYZ to RGB");
  static const bool a_ycbcr2rgb   = cimg_option("-ycbcr2rgb",false,"from YCbCr to RGB");
  static const bool a_yuv2rgb     = cimg_option("-yuv2rgb",false,"from YUV to RGB");
  static const bool a_hsi2rgb     = cimg_option("-hsi2rgb",false,"from HSI to RGB");
  static const bool a_hsl2rgb     = cimg_option("-hsl2rgb",false,"from HSL to RGB");
  static const bool a_hsv2rgb     = cimg_option("-hsv2rgb",false,"from HSV to RGB");

  if (display_help) cimg_help("\n Parameters : Geometric manipulation"
                              "\n -----------------------------------");
  static const char *a_fillpos  = cimg_option("-fill",(char*)0,"Flood fill : Starting position 'x,y,z'");
  static const char *a_fillval  = cimg_option("-vfill","0","Flood fill : Filling value");
  static const float a_filltol  = cimg_option("-tfill",20.0f,"Flood fill : Filling tolerance");
  static const char *a_crop     = cimg_option("-crop",(char*)0,"Crop image(s) (ex : '0,0x100,100' or '0,0,0x50%,50%,2')");
  static const float a_rotate   = cimg_option("-angle",0.0f,"Rotate image(s) (angle in degrees)");
  static const char a_mirror    = cimg_option("-mirror",'\0',"Mirror image(s) (along axis 'x','y','z' or 'v')");
  static const bool a_scale4x   = cimg_option("-s4x",false,"Resize image(s) using 'Scale4x' algorithm");
  static const bool a_scale3x   = cimg_option("-s3x",false,"Resize image(s) using 'Scale3x' algorithm");
  static const bool a_scale2x   = cimg_option("-s2x",false,"Resize image(s) using 'Scale2x' algorithm");
  static const char *a_resize   = cimg_option("-g",(char*)0,"Resize image(s) (ex : '100x100' or '50%x30%')");
  static const int a_interp     = cimg_option("-interp",3,"Resizing interpolation mode "
                                              "(0=none, 1=bloc, 2=moving average, 3=linear, 4=grid, 5=cubic");
  static const int a_bordercond = cimg_option("-bc",-1,"Resizing border condition "
                                              "(-1=none, 0=black, 1=neumann, 2=mosaic)");
  static const bool a_center    = cimg_option("-center",false,"Resizing center flag");
  static const bool a_transpose = cimg_option("-transpose",false,"Transpose image(s)");
  static const char *a_permute  = cimg_option("-permute",(char*)0,"Permute axes of image(s) (ex : 'yzvx')");


  if (display_help) cimg_help("\n Parameters : Image filtering"
                              "\n ----------------------------");
  static const float a_blur   = cimg_option("-blur",0.0f,"Apply gaussian blur (std)");
  static const char *a_smooth = cimg_option("-smooth",(char*)0,"Apply anisotropic diffusion "
                                            "(strength,sharpness,anisotropy,alpha,sigma)");
  static const char *a_bf     = cimg_option("-bf",(char*)0,"Apply bilateral filter (sigma_s,sigma_r)");
  static const int a_median   = cimg_option("-median",0,"Apply median filter (size)");
  static const int a_erode    = cimg_option("-erode",0,"Erode image(s) (size)");
  static const int a_dilate   = cimg_option("-dilate",0,"Dilate image(s) (size)");
  static const float a_noiseg = cimg_option("-ng",0.0f,"Add gaussian noise (std)");
  static const float a_noiseu = cimg_option("-nu",0.0f,"Add uniform noise (std)");
  static const float a_noises = cimg_option("-ns",0.0f,"Add salt&pepper noise (percentage)");

  if (display_help) cimg_help("\n Parameters : Arithmetic operators"
                              "\n ---------------------------------");
  static const bool a_diff = cimg_option("-diff",false,"Compute difference(s) with the first image");
  static const float a_mul = cimg_option("-mul",1.0f,"Multiply image(s) by a float constant");
  static const float a_add = cimg_option("-add",0.0f,"Add a float constant to image(s)");
  static const bool a_abs  = cimg_option("-abs",false,"Compute absolute value");
  static const float a_pow = cimg_option("-pow",1.0f,"Compute power");
  static const bool a_log  = cimg_option("-log",false,"Apply logarithm function");
  static const bool a_sin  = cimg_option("-sin",false,"Apply sine function");
  static const bool a_cos  = cimg_option("-cos",false,"Apply cosine function");

  if (display_help) cimg_help("\n Parameters : Pixel values transformation"
                              "\n ----------------------------------------");
  static const bool a_norm        = cimg_option("-norm",false,"Compute vector norms");
  static const bool a_orient      = cimg_option("-orient",false,"Compute vector orientations");
  static const char *a_cut        = cimg_option("-cut",(char*)0,"Cut pixel values (ex : '0,255')");
  static const char *a_vnorm      = cimg_option("-val",(char*)0,"Normalize pixel values (ex : '0,255')");
  static const int a_quantize     = cimg_option("-quantize",0,"Quantize pixel values (number of levels)");
  static const bool a_equalize    = cimg_option("-equalize",false,"Equalize histogram");
  static const bool a_falsecolors = cimg_option("-falsecolors",false,"Map to false-colors");
  static const bool a_clusters    = cimg_option("-clusters",false,"Map to cluster colors");

  if (display_help) cimg_help("\n Parameters : Output data"
                              "\n ------------------------");
  static const char *a_annot     = cimg_option("-annotate",(char*)0,"Annotate image(s)");
  static const int a_histo       = cimg_option("-histogram",0,"Compute histogram (number of levels)");
  static const char *a_break     = cimg_option("-b",(char*)0,"Break image(s) along axis (can be 'z', 'x3', 'y2', etc...)");
  static const bool a_reverse    = cimg_option("-r",false,"Reverse image list");
  static const char a_append     = cimg_option("-a",'\0',"Append images (along axis 'x','y','z' or 'v')");
  static const char a_unroll     = cimg_option("-unroll",'\0',"Unroll image(s) (along axis 'x','y','z' or 'v')");
  static const bool a_psnr       = cimg_option("-psnr",false,"Compute PSNR with the first image");
  static const char *a_iso       = cimg_option("-iso",(char*)0,"Extract isocurve/isosurface with given value and output object file");
  if (display_help) cimg_option("-t",typei,"Set output type of image pixels (char,uchar,ushort,short,uint,int,ulong,long,float,double)");
  static const char *file_o      = cimg_option("-o",(char*)0,"Set output filename");
  static const char *a_visu      = cimg_option("-visu",(char*)0,"Visualize image(s)");
  static const bool a_plot       = cimg_option("-plot",false,"Visualize as a function plot");

  if (display_help) return;

  if (!src.size) {
    std::fprintf(stderr,"\n**\n** inrcast : You must specify at least one valid input image"
                 "\n** Try '%s -h' to get the list of options\n**\n",cimg::basename(argv[0]));
    std::exit(0);
  }

  // Print image dimensions and exit
  if (a_dimensions) { std::printf("%d %d %d %d",src[0].dimx(),src[0].dimy(),src[0].dimz(),src[0].dimv()); std::exit(0); }

  // Display input image statistics
  std::fprintf(stderr,"\n- Input image(s) :\n");
  cimglist_for(src,l) {
    char fstring[256] = { 0 };
    std::sprintf(fstring," '%s' [%d]",cimg::basename(filename(l).ptr()),l);
    src[l].print(fstring);
  }

  // Invert endianness of pixel buffer
  if (a_endian) {
    std::fprintf(stderr,"\n- Invert endianness of pixel buffer."); std::fflush(stderr);
    cimglist_for(src,l) cimg::invert_endianness(src[l].ptr(),src[l].size());
  }

  // RGB to HSV
  if (a_rgb2hsv) {
    std::fprintf(stderr,"\n- Color space conversion from RGB to HSV."); std::fflush(stderr);
    cimglist_for(src,l) src[l].RGBtoHSV();
  }

  // RGB to HSL
  if (a_rgb2hsl) {
    std::fprintf(stderr,"\n- Color space conversion from RGB to HSL."); std::fflush(stderr);
    cimglist_for(src,l) src[l].RGBtoHSL();
  }

  // RGB to HSI
  if (a_rgb2hsi) {
    std::fprintf(stderr,"\n- Color space conversion from RGB to HSI."); std::fflush(stderr);
    cimglist_for(src,l) src[l].RGBtoHSI();
  }

  // RGB to YUV
  if (a_rgb2yuv) {
    std::fprintf(stderr,"\n- Color space conversion from RGB to YUV."); std::fflush(stderr);
    cimglist_for(src,l) src[l].RGBtoYUV();
  }

  // RGB to YCbCr
  if (a_rgb2ycbcr) {
    std::fprintf(stderr,"\n- Color space conversion from RGB to YCbCr."); std::fflush(stderr);
    cimglist_for(src,l) src[l].RGBtoYCbCr();
  }

  // RGB to XYZ
  if (a_rgb2xyz) {
    std::fprintf(stderr,"\n- Color space conversion from RGB to XYZ."); std::fflush(stderr);
    cimglist_for(src,l) src[l].RGBtoXYZ();
  }

  // RGB to Lab
  if (a_rgb2lab) {
    std::fprintf(stderr,"\n- Color space conversion from RGB to Lab."); std::fflush(stderr);
    cimglist_for(src,l) src[l].RGBtoLab();
  }

  // RGB to CMY
  if (a_rgb2cmy) {
    std::fprintf(stderr,"\n- Color space conversion from RGB to CMY."); std::fflush(stderr);
    cimglist_for(src,l) src[l].RGBtoCMY();
  }

  // RGB to CMYK
  if (a_rgb2cmyk) {
    std::fprintf(stderr,"\n- Color space conversion from RGB to CMYK."); std::fflush(stderr);
    cimglist_for(src,l) src[l].RGBtoCMYK();
  }

  // CMYK to RGB
  if (a_cmyk2rgb) {
    std::fprintf(stderr,"\n- Color space conversion from CMYK to RGB."); std::fflush(stderr);
    cimglist_for(src,l) src[l].CMYKtoRGB();
  }

  // CMY to RGB
  if (a_cmy2rgb) {
    std::fprintf(stderr,"\n- Color space conversion from CMY to RGB."); std::fflush(stderr);
    cimglist_for(src,l) src[l].CMYtoRGB();
  }

  // Lab to RGB
  if (a_lab2rgb) {
    std::fprintf(stderr,"\n- Color space conversion from Lab to RGB."); std::fflush(stderr);
    cimglist_for(src,l) src[l].LabtoRGB();
  }

  // XYZ to RGB
  if (a_xyz2rgb) {
    std::fprintf(stderr,"\n- Color space conversion from XYZ to RGB."); std::fflush(stderr);
    cimglist_for(src,l) src[l].XYZtoRGB();
  }

  // YCbCr to RGB
  if (a_ycbcr2rgb) {
    std::fprintf(stderr,"\n- Color space conversion from YCbCr to RGB."); std::fflush(stderr);
    cimglist_for(src,l) src[l].YCbCrtoRGB();
  }

  // YUV to RGB
  if (a_yuv2rgb) {
    std::fprintf(stderr,"\n- Color space conversion from YUV to RGB."); std::fflush(stderr);
    cimglist_for(src,l) src[l].YUVtoRGB();
  }

  // HSI to RGB
  if (a_hsi2rgb) {
    std::fprintf(stderr,"\n- Color space conversion from HSI to RGB."); std::fflush(stderr);
    cimglist_for(src,l) src[l].HSItoRGB();
  }

  // HSL to RGB
  if (a_hsl2rgb) {
    std::fprintf(stderr,"\n- Color space conversion from HSL to RGB."); std::fflush(stderr);
    cimglist_for(src,l) src[l].HSLtoRGB();
  }

  // HSL to RGB
  if (a_hsv2rgb) {
    std::fprintf(stderr,"\n- Color space conversion from HSV to RGB."); std::fflush(stderr);
    cimglist_for(src,l) src[l].HSVtoRGB();
  }

  // Apply Flood fill algorithm
  if (a_fillpos) {
    std::fprintf(stderr,"\n- Apply flood fill : ");
    char buf[32] = { 0 };
    int cx = 0,cy = 0,cz = 0;
    bool aborted = false;
    if (!std::sscanf(a_fillpos,"%d%15[^0-9-]%d%15[^0-9-]%d%15[^0-9-]",&cx,buf+0,&cy,buf+1,&cz,buf+2)) {
      std::fprintf(stderr,"(Select starting point). "); std::fflush(stderr);
      CImg<int> selection = src[0].get_coordinates(0);
      if (selection[0]<0 || selection[1]<0 || selection[2]<0) { std::fprintf(stderr,"Aborted.."); aborted = true; }
      else {
        cx = selection[0]; cy = selection[1]; cz = selection[2];
        std::fprintf(stderr,"From (%d,%d,%d) with value '%s' and tolerance '%g'.",
                     cx,cy,cz,a_fillval,a_filltol); std::fflush(stderr);
      }
    }
    if (!aborted) cimglist_for(src,l) {
        CImg<T> &img = src[l], value(img.dim,1,1,1);
        if (buf[0]!='%' && buf[1]!='%' && buf[2]!='%') {
          if (!l) std::fprintf(stderr,"From (%d,%d,%d) with value '%s' and tolerance '%g'.",
                               cx,cy,cz,a_fillval,a_filltol);
        } else {
          if (buf[0]=='%') cx = cx*(img.dimx()-1)/100;
          if (buf[1]=='%') cy = cy*(img.dimy()-1)/100;
          if (buf[2]=='%') cz = cz*(img.dimz()-1)/100;
          std::fprintf(stderr,"\n    [%d] : From (%d,%d,%d) with value '%s' and tolerance '%g'.",
                       l,cx,cy,cz,a_fillval,a_filltol); std::fflush(stderr);
        }
        value.fill(a_fillval,true);
        img.draw_fill(cx,cy,cz,value,a_filltol);
      }
  }

  // Crop image(s)
  if (a_crop) {
    std::fprintf(stderr,"\n- Crop image(s) : ");
    char buf[32] = { 0 };
    int c0,c1,c2,c3,c4,c5,c6,c7,
      nb = std::sscanf(a_crop,"%d%15[^0-9-]%d%15[^0-9-]%d%15[^0-9-]%d%15[^0-9-]"
                       "%d%15[^0-9-]%d%15[^0-9-]%d%15[^0-9-]%d%15[^0-9-]",
                       &c0,buf+0,&c1,buf+1,&c2,buf+2,&c3,buf+3,&c4,buf+4,&c5,buf+5,&c6,buf+6,&c7,buf+7);
    if (!nb) {
      std::fprintf(stderr,"(Select image region). "); std::fflush(stderr);
      CImg<int> selection = src[0].get_coordinates(2);
      if (selection[0]<0 || selection[1]<0 || selection[2]<0 ||
          selection[3]<0 || selection[4]<0 || selection[5]<0) std::fprintf(stderr,"Aborted..");
      else {
        std::fprintf(stderr,"From (%d,%d,%d) to (%d,%d,%d).",
                     selection[0],selection[1],selection[2],
                     selection[3],selection[4],selection[5]);
        cimglist_for(src,l) src[l].crop(selection[0],selection[1],selection[2],selection[3],selection[4],selection[5]);
      }
    } else cimglist_for(src,l) {
      CImg<T> &img = src[l];
      int x0 = 0, y0 = 0, z0 = 0, v0 = 0, x1 = img.dimx()-1, y1 = img.dimy()-1, z1 = img.dimz()-1, v1 = img.dimv()-1;
      switch (nb) {
      case 1: case 2:
        x0 = c0;
        if (buf[0]=='%') x0 = x0*(img.dimx()-1)/100;
        break;
      case 3: case 4:
        x0 = c0; x1 = c1;
        if (buf[0]=='%') x0 = x0*(img.dimx()-1)/100;
        if (buf[1]=='%') x1 = x1*(img.dimx()-1)/100;
        break;
      case 5: case 6:
        x0 = c0; y0 = c1; x1 = c2;
        if (buf[0]=='%') x0 = x0*(img.dimx()-1)/100;
        if (buf[1]=='%') y0 = y0*(img.dimy()-1)/100;
        if (buf[2]=='%') x1 = x1*(img.dimx()-1)/100;
        break;
      case 7: case 8:
        x0 = c0; y0 = c1; x1 = c2; y1 = c3;
        if (buf[0]=='%') x0 = x0*(img.dimx()-1)/100;
        if (buf[1]=='%') y0 = y0*(img.dimy()-1)/100;
        if (buf[2]=='%') x1 = x1*(img.dimx()-1)/100;
        if (buf[3]=='%') y1 = y1*(img.dimy()-1)/100;
        break;
      case 9: case 10:
        x0 = c0; y0 = c1; z0 = c2; x1 = c3; y1 = c4;
        if (buf[0]=='%') x0 = x0*(img.dimx()-1)/100;
        if (buf[1]=='%') y0 = y0*(img.dimy()-1)/100;
        if (buf[2]=='%') z0 = z0*(img.dimz()-1)/100;
        if (buf[3]=='%') x1 = x1*(img.dimx()-1)/100;
        if (buf[4]=='%') y1 = y1*(img.dimy()-1)/100;
        break;
      case 11: case 12:
        x0 = c0; y0 = c1; z0 = c2; x1 = c3; y1 = c4; z1 = c5;
        if (buf[0]=='%') x0 = x0*(img.dimx()-1)/100;
        if (buf[1]=='%') y0 = y0*(img.dimy()-1)/100;
        if (buf[2]=='%') z0 = z0*(img.dimz()-1)/100;
        if (buf[3]=='%') x1 = x1*(img.dimx()-1)/100;
        if (buf[4]=='%') y1 = y1*(img.dimy()-1)/100;
        if (buf[5]=='%') z1 = z1*(img.dimz()-1)/100;
        break;
      case 13: case 14:
        x0 = c0; y0 = c1; z0 = c2; v0 = c3; x1 = c4; y1 = c5; z1 = c6;
        if (buf[0]=='%') x0 = x0*(img.dimx()-1)/100;
        if (buf[1]=='%') y0 = y0*(img.dimy()-1)/100;
        if (buf[2]=='%') z0 = z0*(img.dimz()-1)/100;
        if (buf[3]=='%') v0 = v0*(img.dimv()-1)/100;
        if (buf[4]=='%') x1 = x1*(img.dimx()-1)/100;
        if (buf[5]=='%') y1 = y1*(img.dimy()-1)/100;
        if (buf[6]=='%') z1 = z1*(img.dimz()-1)/100;
        break;
      default:
        x0 = c0; y0 = c1; z0 = c2; v0 = c3; x1 = c4; y1 = c5; z1 = c6; v1 = c7;
        if (buf[0]=='%') x0 = x0*(img.dimx()-1)/100;
        if (buf[1]=='%') y0 = y0*(img.dimy()-1)/100;
        if (buf[2]=='%') z0 = z0*(img.dimz()-1)/100;
        if (buf[3]=='%') v0 = v0*(img.dimv()-1)/100;
        if (buf[4]=='%') x1 = x1*(img.dimx()-1)/100;
        if (buf[5]=='%') y1 = y1*(img.dimy()-1)/100;
        if (buf[6]=='%') z1 = z1*(img.dimz()-1)/100;
        if (buf[7]=='%') v1 = z1*(img.dimv()-1)/100;
        break;
      }
      if (x0>x1) cimg::swap(x0,x1);
      if (y0>y1) cimg::swap(y0,y1);
      if (z0>z1) cimg::swap(z0,z1);
      if (v0>v1) cimg::swap(v0,v1);
      std::fprintf(stderr,"\n    [%d] : From (%d,%d,%d,%d) to (%d,%d,%d,%d).",l,x0,y0,z0,v0,x1,y1,z1,v1);
      img.crop(x0,y0,z0,v0,x1,y1,z1,v1);
    }
  }

  // Rotate image(s)
  if (a_rotate!=0) {
    std::fprintf(stderr,"\n- Rotate image(s) with angle '%g' deg.",a_rotate); std::fflush(stderr);
    cimglist_for(src,l) src[l].rotate(a_rotate,3);
  }

  // Mirror image(s)
  if (a_mirror!='\0') {
    std::fprintf(stderr,"\n- Mirror image(s) along axis '%c'.",a_mirror); std::fflush(stderr);
    cimglist_for(src,l) src[l].mirror(a_mirror);
  }

  // ScaleNx image
  if (a_scale4x) {
    std::fprintf(stderr,"\n- Resize image(s) using 'Scale4x' algorithm."); std::fflush(stderr);
    cimglist_for(src,l) src[l].resize_doubleXY().resize_doubleXY();
  }
  if (a_scale3x) {
    std::fprintf(stderr,"\n- Resize image(s) using 'Scale3x' algorithm."); std::fflush(stderr);
    cimglist_for(src,l) src[l].resize_tripleXY();
  }
  if (a_scale2x) {
    std::fprintf(stderr,"\n- Resize image(s) using 'Scale2x' algorithm."); std::fflush(stderr);
    cimglist_for(src,l) src[l].resize_doubleXY();
  }

  // Resize image
  if (a_resize) {
    std::fprintf(stderr,"\n- Resize image(s) : "); std::fflush(stderr);
    char buf[32] = { 0 };
    int dimx = -100, dimy = -100, dimz = -100, dimv = -100,
      nb = std::sscanf(a_resize,"%d%15[^0-9]%d%15[^0-9]%d%15[^0-9]%d%15[^0-9]",
                       &dimx,buf+0,&dimy,buf+1,&dimz,buf+2,&dimv,buf+3);
    if (!nb) {
      std::fprintf(stderr,"Select window size.\t"); std::fflush(stderr);
      CImgDisplay disp(src[0],"Select window size");
      src[0].get_coordinates(0,disp);
      dimx = disp.window_dimx();
      dimy = disp.window_dimy();
      disp.assign();
    } else {
      if (buf[0]=='%') dimx=-dimx;
      if (buf[1]=='%') dimy=-dimy;
      if (buf[2]=='%') dimz=-dimz;
      if (buf[3]=='%') dimv=-dimv;
    }
    cimglist_for(src,l) {
      src[l].resize(dimx,dimy,dimz,dimv,a_interp,a_bordercond,a_center);
      std::fprintf(stderr,"\n    [%d] : Resize to '%dx%dx%dx%d'",l,src[l].dimx(),src[l].dimy(),src[l].dimz(),src[l].dimv());
    }
  }

  // Transpose image
  if (a_transpose) {
    std::fprintf(stderr,"\n- Transpose image(s)."); std::fflush(stderr);
    cimglist_for(src,l) src[l].transpose();
  }

  // Permute image axes
  if (a_permute) {
    std::fprintf(stderr,"\n- Permute axes of image(s) with permutation '%s'.",a_permute); std::fflush(stderr);
    cimglist_for(src,l) src[l].permute_axes(a_permute);
  }

  // Image blur
  if (a_blur!=0) {
    std::fprintf(stderr,"\n- Apply gaussian blur with standard deviation '%g'.",a_blur); std::fflush(stderr);
    cimglist_for(src,l) src[l].blur(a_blur);
  }

  // Apply anisotropic diffusion
  if (a_smooth>0) {
    float strength = 0.0f, sharpness = 0.7f, anisotropy = 0.3f, alpha = 0.6f, sigma = 1.0f;
    std::sscanf(a_smooth,"%f%*c%f%*c%f%*c%f%*c%f",&strength,&sharpness,&anisotropy,&alpha,&sigma);
    std::fprintf(stderr,"\n- Apply anisotropic diffusion with strength '%g', sharpness '%g', anisotropy '%g', alpha '%g', sigma '%g'.",
                 strength,sharpness,anisotropy,alpha,sigma); std::fflush(stderr);
    cimglist_for(src,l) src[l].blur_anisotropic(strength,sharpness,anisotropy,alpha,sigma);
  }

  // Apply bilateral filter
  if (a_bf>0) {
    float sigma_s = 0.0f, sigma_r = 0.0f;
    std::sscanf(a_bf,"%f%*c%f",&sigma_s,&sigma_r);
    std::fprintf(stderr,"\n- Apply bilateral filter with sigma_s '%g', sigma_r '%g'.",sigma_s,sigma_r); std::fflush(stderr);
    cimglist_for(src,l) src[l].blur_bilateral(sigma_s,sigma_r);
  }

  // Apply median filter
  if (a_median) {
    std::fprintf(stderr,"\n- Apply median filter of size '%d'.",a_median); std::fflush(stderr);
    cimglist_for(src,l) src[l].blur_median(a_median);
  }

  // Erode image(s)
  if (a_erode) {
    std::fprintf(stderr,"\n- Erode image(s) with size '%d'.",a_erode); std::fflush(stderr);
    cimglist_for(src,l) src[l].erode(a_erode);
  }

  // Dilate image(s)
  if (a_dilate) {
    std::fprintf(stderr,"\n- Dilate image(s) with size '%d'.",a_dilate); std::fflush(stderr);
    cimglist_for(src,l) src[l].dilate(a_dilate);
  }

  // Add gaussian noise
  if (a_noiseg) {
    std::fprintf(stderr,"\n- Add gaussian noise with standard deviation '%g'.",a_noiseg); std::fflush(stderr);
    cimglist_for(src,l) src[l].noise(a_noiseg,0);
  }

  // Add uniform noise
  if (a_noiseu) {
    std::fprintf(stderr,"\n- Add uniform noise with standard deviation '%g'.",a_noiseu); std::fflush(stderr);
    cimglist_for(src,l) src[l].noise(a_noiseu,1);
  }

  // Add salt&pepper noise
  if (a_noises) {
    std::fprintf(stderr,"\n- Add Salt&Pepper noise with percentage '%g'.",a_noises); std::fflush(stderr);
    cimglist_for(src,l) src[l].noise(a_noises,2);
  }

  // Compute difference with the first image
  if (a_diff) {
    std::fprintf(stderr,"\n- Compute difference(s) with the first image."); std::fflush(stderr);
    for (int l = src.size-1; l>=0; --l) src[l]-=src[0];
  }

  // Multiplication image(s) by a float constant
  if (a_mul!=1.0f) {
    std::fprintf(stderr,"\n- Multiply image(s) by '%g'.",a_mul); std::fflush(stderr);
    cimglist_for(src,l) src[l]*=a_mul;
  }

  // Add a float constant to image(s)
  if (a_add!=0.0f) {
    std::fprintf(stderr,"\n- Add '%g' to image(s).",a_add); std::fflush(stderr);
    cimglist_for(src,l) src[l]+=(T)a_add;
  }

  // Compute absolute value
  if (a_abs) {
    std::fprintf(stderr,"\n- Compute absolute values."); std::fflush(stderr);
    cimglist_for(src,l) src[l].abs();
  }

  // Compute power
  if (a_pow!=1.0f) {
    std::fprintf(stderr,"\n- Compute image(s) to the power '%g'.",a_pow); std::fflush(stderr);
    cimglist_for(src,l) src[l].pow(a_pow);
  }

  // Apply logarithm function
  if (a_log) {
    std::fprintf(stderr,"\n- Apply logarithm function."); std::fflush(stderr);
    cimglist_for(src,l) src[l].log();
  }

  // Apply sine function
  if (a_sin) {
    std::fprintf(stderr,"\n- Apply sine function."); std::fflush(stderr);
    cimglist_for(src,l) src[l].sin();
  }

  // Apply cosine function
  if (a_cos) {
    std::fprintf(stderr,"\n- Apply cosine function."); std::fflush(stderr);
    cimglist_for(src,l) src[l].cos();
  }

  // Compute vector norms
  if (a_norm) {
    std::fprintf(stderr,"\n- Compute vector norms."); std::fflush(stderr);
    cimglist_for(src,l) src[l].norm_pointwise();
  }

  // Compute vector orientations
  if (a_orient) {
    std::fprintf(stderr,"\n- Compute vector orientations."); std::fflush(stderr);
    cimglist_for(src,l) src[l].orientation_pointwise();
  }

  // Cut pixel values
  if (a_cut) {
    std::fprintf(stderr,"\n- Cut pixel values in '%s'.",a_cut); std::fflush(stderr);
    cimglist_for(src,l) {
      CImg<T>& img = src[l];
      double vmin, vmax = img.maxmin(vmin);
      std::sscanf(a_cut,"%lg%*c%lg",&vmin,&vmax);
      img.cut((T)vmin,(T)vmax);
    }
  }

  // Normalize pixel values
  if (a_vnorm) {
    std::fprintf(stderr,"\n- Normalize pixel values in '%s'.",a_vnorm); std::fflush(stderr);
    cimglist_for(src,l) {
      CImg<T>& img = src[l];
      double vmin, vmax = img.maxmin(vmin);
      std::sscanf(a_vnorm,"%lg%*c%lg",&vmin,&vmax);
      img.normalize((T)vmin,(T)vmax);
    }
  }

  // Quantize pixel values
  if (a_quantize) {
    std::fprintf(stderr,"\n- Quantize pixel values with '%d' levels.",a_quantize); std::fflush(stderr);
    cimglist_for(src,l) src[l].quantize(a_quantize);
  }

  // Equalize histogram
  if (a_equalize) {
    std::fprintf(stderr,"\n- Equalize histogram."); std::fflush(stderr);
    cimglist_for(src,l) src[l].equalize_histogram(4096);
  }

  // Map to false-colors
  if (a_falsecolors) {
    std::fprintf(stderr,"\n- Map to false-colors."); std::fflush(stderr);
    cimglist_for(src,l) {
      CImg<unsigned char> tmp = src[l].get_normalize((T)0,(T)255);
      cimg_forXYZV(tmp,x,y,z,k) src[l](x,y,z,k) = (T)cimg::ror((unsigned char)tmp(x,y,z,k),4);
    }
  }

  // Map to cluster colors
  if (a_clusters) {
    std::fprintf(stderr,"\n- Map to cluster colors."); std::fflush(stderr);
    const CImg<T> palette = CImg<T>::get_contrast_LUT8();
    cimglist_for(src,l) src[l].norm_pointwise(1).normalize(0,(T)255).LUTtoRGB(palette);
  }

  // Annotate images
  if (a_annot) {
    std::fprintf(stderr,"\n- Annotate image(s) with '%s'.",a_annot); std::fflush(stderr);
    cimglist_for(src,l) {
      const CImg<T> col(src[l].dimv(),1,1,1,src[l].max());
      src[l].draw_text(a_annot,5,5,col.ptr());
    }
  }

  // Compute image histogram
  if (a_histo) {
    std::fprintf(stderr,"\n- Compute histogram(s) with '%d' levels.",a_histo); std::fflush(stderr);
    cimglist_for(src,l) src[l] = src[l].get_histogram(a_histo);
  }

  // Break image(s)
  if (a_break) {
    int nb = 0;
    std::sscanf(a_break,"%*[xyzvXYZV]%d",&nb);
    std::fprintf(stderr,"\n- Break image(s) along axis '%c'.",a_break[0]); std::fflush(stderr);
    CImgList<T> tmp;
    cimglist_for(src,l) tmp.insert(src[l].get_split(a_break[0],nb));
    src = tmp;
  }

  // Reverse image list
  if (a_reverse) {
    std::fprintf(stderr,"\n- Reverse image list."); std::fflush(stderr);
    src.reverse();
  }

  // Append image(s)
  if (a_append!='\0') {
    std::fprintf(stderr,"\n- Append image(s) along axis '%c'.",a_append); std::fflush(stderr);
    src = CImgList<T>(src.get_append(a_append,'p'));
  }

  // Unroll image(s)
  if (a_unroll) {
    std::fprintf(stderr,"\n- Unroll image(s) along axis '%c'.",a_unroll); std::fflush(stderr);
    cimglist_for(src,l) src[l].unroll(a_unroll);
  }

  // Compute PSNR
  if (a_psnr) {
    std::fprintf(stderr,"\n- Compute PSNR with the first image : "); std::fflush(stderr);
    cimglist_for(src,l)
      if (src[0].is_sameXYZV(src[l])) std::fprintf(stderr,"\n    [%d] : PSNR(0,%u) = %g\n",l,l,src[0].PSNR(src[l]));
  }

  // Find isocurve/isosurface and output 3D object file(s)
  if (a_iso) {
    const CImg<unsigned char> default_color = CImg<unsigned char>::vector(190,190,200);
    float isovalue = 0;
    std::sscanf(a_iso,"%f",&isovalue);
    std::fprintf(stderr,"\n- Compute isocurve/isosurface with value '%g' : ",isovalue); std::fflush(stderr);
    cimglist_for(src,l) {
      CImgList<> points;
      CImgList<unsigned int> primitives;
      if (src[l].depth>1) {
        src[l].marching_cubes(isovalue,points,primitives);
        const bool ifaces = cimg_option("-if",false,0);
        if (ifaces) cimglist_for(primitives,p) cimg::swap(primitives(p,0),primitives(p,1));
        std::fprintf(stderr,"\n    [%d] : Surface with %d vertices, %d triangles.\n",l,points.size,primitives.size);
      } else {
        src[l].marching_squares(isovalue,points,primitives);
        std::fprintf(stderr,"\n    [%d] : Curve with %d vertices, %d segments.\n",l,points.size,primitives.size);
      }
      const CImgList<unsigned char> colors(primitives.size,default_color,true);
      if (file_o) {
        char tmp[512] = { 0 };
        if (src.size!=1) cimg::filename_number(file_o,l,6,tmp); else std::sprintf(tmp,"%s",file_o);
        points.save_off(tmp,primitives,colors);
      }
      if (a_visu || !file_o) {
        const char *const bgcolor = cimg_option("-bgcolor","32,64,96",0);
        const int render_static  = cimg_option("-render",4,0);
        const int render_motion  = cimg_option("-renderm",-1,0);
        const bool ofaces        = cimg_option("-of",false,0);
        unsigned int colr = 0, colg = 0, colb = 0;
        std::sscanf(bgcolor,"%u%*c%u%*c%u",&colr,&colg,&colb);
        static CImgDisplay disp(512,512,cimg::basename(filename(l).ptr()),0);
        static const CImg<unsigned char> background = CImg<unsigned char>(3,disp.dimx(),disp.dimy(),1).
          fill((unsigned char)colr,(unsigned char)colg,(unsigned char)colb).permute_axes("yzvx");
        if (!disp.is_closed) {
          disp.set_title("%s (isovalue '%g')",cimg::basename(filename(l).ptr()),isovalue);
          background.display_object3d(points,primitives,colors,disp,true,render_static,render_motion,!ofaces,
                                      500.0f,0.7f,0.3f);
        }
      }
    }
    return;
  }

  // Set output type of image pixels
  std::fprintf(stderr,"\n- Set output type of image pixels to '%s'.\t",cimg::type<t>::string()); std::fflush(stderr);
  CImgList<t> dest(src);
  src.assign();
  dest.print("\n- Output images");

  // Save output image list
  if (file_o) {
    std::fprintf(stderr,"\n- Save image(s) to '%s'.\t",file_o); std::fflush(stderr);
    dest.save(file_o);
  }

  // Display destination image
  if ((!a_visu && !file_o) || (a_visu &&
                               cimg::strncasecmp(a_visu,"off",3) &&
                               cimg::strncasecmp(a_visu,"0",1) &&
                               cimg::strncasecmp(a_visu,"false",5))) {
    if (dest.size==1 && a_plot) {
      std::fprintf(stderr,"\n- Visualize image as a function plot.\t"); std::fflush(stderr);
      CImg<t> resized_img;
      const CImg<t>& img = dest[0].dimv()<=8?dest[0]:(resized_img=dest[0].get_resize(-100,-100,-100,8,0));
      CImg<unsigned char> palette = CImg<unsigned char>(3,img.dimv(),1,1,0).noise(300,1);
      const unsigned char gray[3]={128,128,128},white[3]={255,255,255};
      unsigned long trame = 0x0F0F0F0F;
      char message[1024];
      palette[0]=255; palette[1]=0; palette[2]=0;
      if (img.dimv()>1) { palette[3]=0; palette[4]=255; palette[5]=0; }
      if (img.dimv()>2) { palette[6]=0; palette[7]=0; palette[8]=255; }
      CImgDisplay disp(500,300,"1D plot of the output image",0);
      CImg<unsigned char> visu0;
      do {
        if (!visu0.ptr() || disp.is_resized) { // plot graphics
          disp.resize();
          visu0 = CImg<unsigned char>(disp.dimx(),disp.dimy(),1,3,0);
          if (img.dimv()==1) visu0.draw_graph(img,palette,2,img.max(),img.min(),0.4f);
          cimg_forV(img,k) visu0.draw_graph(img.get_shared_channel(k),&palette(0,k),1,img.max(),img.min());
          visu0.draw_axis(CImg<unsigned int>::sequence(1+visu0.dimx()/60,0,(int)img.size()-1),
                          CImg<double>::sequence(1+visu0.dimy()/60,img.max(),img.min()),
                          gray);
        }
        CImg<unsigned char> visu(visu0);
        if (disp.mouse_x>=0) {
          visu.draw_line(disp.mouse_x,0,disp.mouse_x,visu.dimy()-1,gray,0.5f,trame);
          trame = cimg::rol(trame);
          const unsigned x = disp.mouse_x*img.size()/(img.dimv()*disp.dimx());
          std::sprintf(message,"x=%d\n[ ",x);
          cimg_forV(img,k) std::sprintf(message+std::strlen(message),"%s%g%s",k==0?"":"  ",
                                        (float)img(x,0,0,k),k==img.dimv()-1?"":"\n");
          std::sprintf(message+std::strlen(message)," %s]",dest[0].dimv()<=8?"":"\n...");
          visu.draw_text(message,disp.mouse_x-20,10,white,0,1);
        } else visu=visu0;
        disp.display(visu).wait(40);
      } while (!disp.is_closed && !disp.key);
    } else {
      std::fprintf(stderr,"\n- Visualize image(s) : Use mouse/keyboard to zoom on regions.\t"); std::fflush(stderr);
      char win_title[256] = { 0 };
      if (dest.size>1) std::sprintf(win_title,"%s... [#%d]",cimg::basename(filename(0).ptr()),dest.size);
      else std::sprintf(win_title,"%s",cimg::basename(filename(0).ptr()));
      dest.display(win_title,3,false);
    }
    std::fprintf(stderr,"\n");
  }
}

//---------------------------------------------------------
// Read input image and do conversion to output pixel type
//---------------------------------------------------------
template<typename T> void input_file(int argc, char **argv, T& pixel_type) {
  pixel_type = 0;

  // Read input image list
  CImgList<T> src;
  CImgList<char> filename;
  output_file(argc,argv,src,filename,true,pixel_type);  // just to display help
  std::fprintf(stderr,"\n- Load input image(s)...\t"); std::fflush(stderr);

  bool opt = false;
  int foo = 0;
  for (int k = 1; k<argc; k++)
    if (argv[k][0]=='-' && argv[k][1]!='\0' && argv[k][1]!='.') opt = true;
    else {
      if (!opt ||
          !cimg::strcasecmp(argv[k-1],"-h") ||
          !cimg::strcasecmp(argv[k-1],"-orient") ||
          !cimg::strcasecmp(argv[k-1],"-norm") ||
          !cimg::strcasecmp(argv[k-1],"-equalize") ||
          !cimg::strcasecmp(argv[k-1],"-falsecolors") ||
          (!cimg::strcasecmp(argv[k-1],"-crop") && !std::sscanf(argv[k],"%d",&foo)) ||
          (!cimg::strcasecmp(argv[k-1],"-g") && !std::sscanf(argv[k],"%d",&foo)) ||
          (!cimg::strcasecmp(argv[k-1],"-fpos") && !std::sscanf(argv[k],"%d",&foo)) ||
          !cimg::strcasecmp(argv[k-1],"-get_size") ||
          !cimg::strcasecmp(argv[k-1],"-get_width") ||
          !cimg::strcasecmp(argv[k-1],"-get_height") ||
          !cimg::strcasecmp(argv[k-1],"-get_depth") ||
          !cimg::strcasecmp(argv[k-1],"-get_dim")
          ) {
        std::fprintf(stderr,"\n  > Loading [%d]='%s'\t",k-1,argv[k]); std::fflush(stderr);

        CImgList<T> tmpl;
        const char *ext = cimg::split_filename(argv[k]);
        bool checked = false;

        if (!cimg::strcasecmp(ext,"yuv")) { // YUV sequence file
          unsigned int w = 352, h = 288;
          const char *geom = cimg_option("-g0","352x288",0);
          if (geom) std::sscanf(geom,"%u%*c%u",&w,&h);
          int first_frame = 0, last_frame = ~0U, step_frame = 1;
          const char *frames = cimg_option("-frames",(char*)0,0);
          if (frames) std::sscanf(frames,"%d%*c%d%*c%d",&first_frame,&last_frame,&step_frame);
          const unsigned oexc = cimg::exception_mode(); cimg::exception_mode() = 0;
          try { tmpl.load_yuv(argv[k],w,h,(unsigned int)first_frame,(unsigned int)last_frame,(unsigned int)step_frame,true); }
          catch (CImgException& e) { std::fprintf(stderr,"\n**\n** inrcast : File '%s' cannot be opened and will be ignored.\n**\n",argv[k]); }
          cimg::exception_mode() = oexc;
          checked = true;
        }

        if (!cimg::strcasecmp(ext,"avi") || // Generic video sequence
            !cimg::strcasecmp(ext,"mov") ||
            !cimg::strcasecmp(ext,"mpg") ||
            !cimg::strcasecmp(ext,"mpeg") ||
            !cimg::strcasecmp(ext,"ogg") ||
            !cimg::strcasecmp(ext,"flv")) {
          int first_frame = 0, last_frame = ~0U, step_frame = 1;
          const char *frames = cimg_option("-frames",(char*)0,0);
          if (frames) std::sscanf(frames,"%d%*c%d%*c%d",&first_frame,&last_frame,&step_frame);
          const unsigned oexc = cimg::exception_mode(); cimg::exception_mode() = 0;
          try { tmpl.load_ffmpeg(argv[k],(unsigned int)first_frame,(unsigned int)last_frame,(unsigned int)step_frame); }
          catch (CImgException& e) { std::fprintf(stderr,"\n**\n** inrcast : File '%s' cannot be opened and will be ignored.\n**\n",argv[k]); }
          cimg::exception_mode() = oexc;
          checked = true;
        }

        if (!cimg::strcasecmp(ext,"raw")) { // Raw data buffer
          unsigned int w = 0, h = 1, d = 1, v = 1;
          const char *geom = cimg_option("-g0",(char*)0,0);
          if (geom) std::sscanf(geom,"%u%*c%u%*c%u%*c%u",&w,&h,&d,&v);
          else {
            std::fprintf(stderr,"\n**\n** inrcast : You must specify an input geometry (option -g0) for .raw images\n**\n");
            std::exit(0);
          }
          const bool multiplexed = cimg_option("-multiplexed",false,0);
          const bool endian_swap = cimg_option("-endian",false,0);
          const unsigned oexc = cimg::exception_mode(); cimg::exception_mode() = 0;
          try { tmpl = CImgList<T>(CImg<T>::get_load_raw(argv[k],w,h,d,v,multiplexed,endian_swap)); }
          catch (CImgException& e) { std::fprintf(stderr,"\n**\n** inrcast : File '%s' cannot be opened and will be ignored.\n**\n",argv[k]); }
          cimg::exception_mode() = oexc;
          checked = true;
        }

        if (!cimg::strcasecmp(ext,"rgb") || !cimg::strcasecmp(ext,"rgba")) { // RGB data buffer
          cimg_help("\n");
          unsigned int w = 0, h = 1;
          const char *geom = cimg_option("-g0",(char*)0,0);
          if (geom) std::sscanf(geom,"%u%*c%u",&w,&h);
          else {
            std::fprintf(stderr,"\n**\n** inrcast : You must specify an input geometry (option -g0) for .rgb or .rgba images\n**\n");
            std::exit(0);
          }
          const unsigned oexc = cimg::exception_mode(); cimg::exception_mode() = 0;
          try {
            if (!cimg::strcasecmp(ext,"rgb")) tmpl = CImgList<T>(CImg<T>::get_load_rgb(argv[k],w,h));
            else tmpl = CImgList<T>(CImg<T>::get_load_rgba(argv[k],w,h));
          }
          catch (CImgException& e) { std::fprintf(stderr,"\n**\n** inrcast : File '%s' cannot be opened and will be ignored.\n**\n",argv[k]); }
          cimg::exception_mode() = oexc;
          checked = true;
        }

        if (!cimg::strcasecmp(ext,"cimg")) { // CIMG hyperspectral and volumetric image list.
          int first_frame = 0, last_frame = ~0U, step_frame = 1;
          const char *geom = cimg_option("-g0",(char*)0,0);
          if (geom) {
            std::sscanf(geom,"%d%*c%d%*c%d",&first_frame,&last_frame,&step_frame);
            const unsigned oexc = cimg::exception_mode(); cimg::exception_mode() = 0;
            try { tmpl = CImgList<T>::get_load_cimg(argv[k],(unsigned int)first_frame,(unsigned int)last_frame,0,0,0,0,~0U,~0U,~0U,~0U); }
            catch (CImgException& e) { std::fprintf(stderr,"\n**\n** inrcast : File '%s' cannot be opened and will be ignored.\n**\n",argv[k]); }
            cimg::exception_mode() = oexc;
          } else {
            const unsigned oexc = cimg::exception_mode(); cimg::exception_mode() = 0;
            try { tmpl = CImgList<T>(argv[k]); }
            catch (CImgException& e) { std::fprintf(stderr,"\n**\n** inrcast : File '%s' cannot be opened and will be ignored.\n**\n",argv[k]); }
            cimg::exception_mode() = oexc;
          }
          checked = true;
        }

        if (!cimg::strcasecmp(ext,"off")) { // OFF 3D objects
          const char *const bgcolor = cimg_option("-bgcolor","32,64,96",0);
          const int render_static  = cimg_option("-render",4,0);
          const int render_motion  = cimg_option("-renderm",-1,0);
          const bool ofaces        = cimg_option("-of",false,0);
          const bool ifaces        = cimg_option("-if",false,0);
          unsigned int colr = 0, colg = 0, colb = 0;
          std::sscanf(bgcolor,"%u%*c%u%*c%u",&colr,&colg,&colb);
          static CImgDisplay disp(512,512,cimg::basename(argv[k]),0);
          const CImg<unsigned char> background = CImg<unsigned char>(3,disp.dimx(),disp.dimy(),1).
            fill((unsigned char)colr,(unsigned char)colg,(unsigned char)colb).permute_axes("yzvx");
          if (!disp.is_closed) {
            disp.set_title(cimg::basename(argv[k]));
            CImgList<unsigned char> colors;
            CImgList<unsigned int> primitives;
            CImg<float> points;
            const unsigned oexc = cimg::exception_mode(); cimg::exception_mode() = 0;
            try { points = CImg<float>::get_load_off(argv[k],primitives,colors,ifaces); }
            catch (CImgException& e) { std::fprintf(stderr,"\n**\n** inrcast : File '%s' cannot be opened and will be ignored.\n**\n",argv[k]); }
            cimg::exception_mode() = oexc;
            std::fprintf(stderr,"\n    ( %u vertices, %u primitives ).",points.dimx(),primitives.size);
            std::fflush(stderr);
            background.display_object3d(points,primitives,colors,disp,true,render_static,render_motion,!ofaces,
                                        500.0f,0.7f,0.3f);
          } else std::exit(0);
          checked = true;
        }

        if (!checked) {
          const unsigned oexc = cimg::exception_mode(); cimg::exception_mode() = 0;
          try { tmpl = CImgList<T>(argv[k]); }
          catch (CImgException& e) { std::fprintf(stderr,"\n**\n** inrcast : File '%s' cannot be opened and will be ignored.\n**\n",argv[k]); }
          cimg::exception_mode() = oexc;
        }

        const CImg<char> name(argv[k],cimg::strlen(argv[k])+1,1,1,1,false);
        if (tmpl.size==1) filename.insert(name);
        else cimglist_for(tmpl,l) {
            char s[1024];
            std::sprintf(s,"%s [%d]",name.ptr(),l);
            filename.insert(CImg<char>(s,cimg::strlen(s)+1,1,1,1,false));
          }

        const unsigned int pos = src.size;
        src.insert(tmpl.size);
        cimglist_for(tmpl,l) tmpl[l].transfer_to(src[pos+l]);
      }
      opt=false;
    }
  std::fputc('\n',stderr);

  // Convert to output pixel type (if specified)
  const char *typei = cimg_option("-i","float",0);
  const char *typeo = cimg_option("-t",typei,0);
  bool saved = false;
  if (!saved && !cimg::strcasecmp(typeo,"bool"))   { bool t = false;       output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"char"))   { char t = 0;           output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"uchar"))  { unsigned char t = 0;  output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"ushort")) { unsigned short t = 0; output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"short"))  { short t = 0;          output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"uint"))   { unsigned int t = 0;   output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"int"))    { int t = 0;            output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"ulong"))  { unsigned long t = 0;  output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"long"))   { long t = 0;           output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"float"))  { float t = 0;          output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved && !cimg::strcasecmp(typeo,"double")) { float t = 0;          output_file(argc,argv,src,filename,false,t); saved = true; }
  if (!saved) {
    std::fprintf(stderr,"\n**\n** inrcast : Cannot save image with pixel type '%s' (see options)\n**\n",typeo);
    std::exit(0);
  }
}

//----------------
// Main procedure
//----------------
int main(int argc,char **argv) {
  cimg_usage("1D/2D/3D Image Converter and Viewer.\n"
             "           usage : inrcast image_in1 [image_in2 [...]] [options]\n"
             "           (processing is done in the order of the following options)");
  if (argc==1) { std::fprintf(stderr,"\n**\n** Try '%s -h' to get the list of options\n**\n",cimg::basename(argv[0])); std::exit(0); }

  // Read input image and call converter
  bool casted = false;
  const char *typei = cimg_option("-i","float",0);
  if (!casted && !cimg::strcasecmp(typei,"bool"))   { bool t = false;       input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"char"))   { char t = 0;           input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"uchar"))  { unsigned char t = 0;  input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"ushort")) { unsigned short t = 0; input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"short"))  { short t = 0;          input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"uint"))   { unsigned int t = 0;   input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"int"))    { int t = 0;            input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"ulong"))  { unsigned long t = 0;  input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"long"))   { long t = 0;           input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"float"))  { float t = 0;          input_file(argc,argv,t); casted=true; }
  if (!casted && !cimg::strcasecmp(typei,"double")) { float t = 0;          input_file(argc,argv,t); casted=true; }
  if (!casted) {
    std::fprintf(stderr,"\n**\n** inrcast : Cannot load image pixel type '%s' (see options)\n**\n",typei);
    std::exit(0);
  }

  // Exit
  std::exit(0);
  return 0;
}
